package thesis.crawler;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLEncoder;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;

import org.htmlcleaner.*;
import org.htmlparser.beans.*;

public class HtmlParser {
	
	public static final String TAG_COMMENTS_REGEX = "<!--(?s).*?-->";
	public static final String URL_REGEX = "[a-zA-z]+://(?s).*?";
	public static final int DOMAIN_LIMITATION = 10;
	
	private static HtmlParser parser;
	private static HtmlCleaner cleaner;
	private static TagNode root;
	
	private HtmlParser()
	{}
	
	public static HtmlParser getInstance()
	{
		if (parser == null)
		{
			parser = new HtmlParser();
			cleaner = new HtmlCleaner();
			root = null;
		}
		return parser;
	}
	
	public void setPageContent(String content) throws IOException
	{
		root = cleaner.clean(content);
	}
	
	/**
	 * Get web page content by removing HTML tags
	 * @param pageUrl The URL string of target web page
	 * @return Pure web page content
	 * @throws IOException 
	 */
	public String getPlainContent() throws IOException
	{
		String url = getUrl();
        StringBean stringBean = new StringBean();
        stringBean.setURL(url);
		return stringBean.getStrings();
	}
	
	public String getUrl() throws IOException
	{
		TagNode urlNode = root.findElementByName("self_url", true);
		return urlNode.getAttributeByName("href");
	}
	
	/**
	 * Get web page title
	 * @param url The URL string of web page
	 * @return Web page title
	 * @throws IOException
	 */
	public String getTitle()
	{
		try
		{
			TagNode head = root.findElementByName("head", true);
			TagNode title = head.findElementByName("title", true);
			return title.getText().toString();
		}
		catch (Exception e)
		{
			return "";
		}
	}
	
	/**
	 * Get web page description
	 * @param url The URL string of web page
	 * @return Web page description/summary/snippet
	 * @throws XPatherException
	 * @throws IOException
	 */
	public String getDescription() throws XPatherException, IOException
	{
		try
		{
		TagNode head = root.findElementByName("head", true);
		Object[] objects = head.evaluateXPath("//meta[@name='description']");
		if (objects.length > 0)
		{
			TagNode description = (TagNode)objects[0];
			return description.getAttributeByName("content");
		}
		return "";
		}
		catch (Exception e)
		{
			return "";
		}
	}
	
	/**
	 * Retrieve the in-links of the destination web page through Google search engine
	 * @param destUrl The URL string of destination web page
	 * @param inlinkNum The Maximum of in-link size of destination web page
	 * @return The in-link array of destination web page
	 * @throws UnknownHostException
	 * @throws IOException
	 * @throws XPatherException
	 * @throws InterruptedException 
	 */
	public ArrayList<String> getInLinks(String destUrl, int inlinkNum) throws UnknownHostException, IOException, XPatherException, InterruptedException
	{
		return GoogleCrawler.getInstance().getInlinkURLs(destUrl, inlinkNum);
	}
	
	public HashMap<String, Integer> getOutlinkDomains(String srcPageUrl, String content) throws IOException, XPatherException
	{
		HtmlCleaner cleaner = new HtmlCleaner();
		String content1 = removeTags(content, TAG_COMMENTS_REGEX);
		TagNode root = cleaner.clean(content1);
		
		HashMap<String, Integer> map = new HashMap<String, Integer>();
		String selfDomain = GoogleCrawler.getInstance().getDomain(srcPageUrl);
		if (selfDomain == null)
			return null;
		else
			map.put(selfDomain, new Integer(1));
		
		Object[] objects = root.evaluateXPath("//a");
		for (int i = 0; i < objects.length; i++)
		{
			TagNode aNode = (TagNode)objects[i];
			String href = aNode.getAttributeByName("href");
			String hrefDomain = GoogleCrawler.getInstance().getDomain(href);
			
			if (hrefDomain == null)
				continue;
			else
			{
				if (!map.containsKey(hrefDomain))
				{
					map.put(hrefDomain, new Integer(1));
				}
				else if (!hrefDomain.equals(selfDomain))
				{
					Integer value = (Integer)map.get(hrefDomain);
					int intValue = value.intValue();
					if (intValue < DOMAIN_LIMITATION)
					{
						intValue++;
						map.put(hrefDomain, new Integer(intValue));
					}
					else
						continue;
				}
				else
					continue;
			}
		}
		
		return map;
	}
	
	/***
	 * 
	 * @param srcPageUrl
	 * @return
	 * @throws IOException
	 * @throws XPatherException
	 */
	public HashMap<String, Integer> getOutlinkDomains(String srcPageUrl) throws IOException, XPatherException
	{
		HtmlCleaner cleaner = new HtmlCleaner();
		String content = GoogleCrawler.getInstance().getPageContent(srcPageUrl);
		content = removeTags(content, TAG_COMMENTS_REGEX);
		TagNode root = cleaner.clean(content);
		
		HashMap<String, Integer> map = new HashMap<String, Integer>();
		String selfDomain = GoogleCrawler.getInstance().getDomain(srcPageUrl);
		if (selfDomain == null)
			return null;
		else
			map.put(selfDomain, new Integer(1));
			
		Object[] objects = root.evaluateXPath("//a");
		for (int i = 0; i < objects.length; i++)
		{
			TagNode aNode = (TagNode)objects[i];
			String href = aNode.getAttributeByName("href");
			String hrefDomain = GoogleCrawler.getInstance().getDomain(href);
			
			if (hrefDomain == null)
				continue;
			else
			//if (href != null && href.matches(URL_REGEX))
			{
				if (!map.containsKey(hrefDomain))
				{
					map.put(hrefDomain, new Integer(1));
				}
				else if (!hrefDomain.equals(selfDomain))
				{
					Integer value = (Integer)map.get(hrefDomain);
					int intValue = value.intValue();
					if (intValue < DOMAIN_LIMITATION)
					{
						intValue++;
						map.put(hrefDomain, new Integer(intValue));
					}
					else
						continue;
				}
				else
					continue;
			}
		}
		
		return map;
	}
	
	public String removeTags(String htmlContent, String tagRegex)
	{
		return htmlContent.replaceAll(tagRegex, "");		
	}
	
	/**
	 * Get Google PageRank of specified domain
	 * @param urlStr The URL string of the domain
	 * @return The Google PageRank
	 * @throws IOException 
	 */
	public int getPageRank(String url) throws IOException
	{
        int pageRank = 0;
        String query = getQueryUrl(url);
        if(query == null)
            return pageRank;
        BufferedReader in = null;
        try {
            URL pr = new URL(query);
            URLConnection conn = pr.openConnection();
            in = new BufferedReader(new InputStreamReader(conn.getInputStream()));
            String line = null;
            do
            {
                if((line = in.readLine()) == null)
                    break;
                if(line.contains(":")) {
                    String tokens[] = line.split(":");
                    if(tokens.length > 2)
                        pageRank = Integer.parseInt(tokens[2]);
                }
            } while(true);
        } catch(Exception e) { }
        return pageRank;
	}
	
    private String getQueryUrl(String url) {
        int checksum;
        checksum = getCheckSum(url);
        if(checksum == 0)
            return null;
        
        String temp = "";
        try {
            
            temp = URLEncoder.encode(url, "UTF-8");
        } catch (UnsupportedEncodingException ex) {
            ex.printStackTrace();
        }
        return "http://www.google.com/search?client=navclient-auto&ch=6" + checksum + "&ie=UTF-8&oe=UTF-8&features=Rank" + "&q=info:" + temp;
    }
	
    private int getCheckSum(String url) {
        if(!validUrl(url))
            return 0;
        else
            return generateCheckSum(strord("info:" + url));
    }
    
    private boolean validUrl(String url) {
        if(url == null || !url.startsWith("http"))
            return false;
        try {
            new URL(url);
        } catch(MalformedURLException e) {
            return false;
        }
        return true;
    }
    
    private int[] strord(String str) {
        int result[] = new int[str.length()];
        for(int i = 0; i < str.length(); i++)
            result[i] = str.charAt(i);
        
        return result;
    }
    
    private static int zeroFill(int a, int b) {
        int z = 0x80000000;
        if((z & a) != 0) {
            a >>= 1;
            a &= ~z;
            a |= 0x40000000;
            a >>= b - 1;
        } else {
            a >>= b;
        }
        return a;
    }
    
    private int[] mix(int a, int b, int c) {
        a -= b;
        a -= c;
        a ^= zeroFill(c, 13);
        b -= c;
        b -= a;
        b ^= a << 8;
        c -= a;
        c -= b;
        c ^= zeroFill(b, 13);
        a -= b;
        a -= c;
        a ^= zeroFill(c, 12);
        b -= c;
        b -= a;
        b ^= a << 16;
        c -= a;
        c -= b;
        c ^= zeroFill(b, 5);
        a -= b;
        a -= c;
        a ^= zeroFill(c, 3);
        b -= c;
        b -= a;
        b ^= a << 10;
        c -= a;
        c -= b;
        c ^= zeroFill(b, 15);
        return (new int[] {
            a, b, c
        });
    }
    
    private int generateCheckSum(int url[]) {
        int length = url.length;
        int init = 0xe6359a60;
        int a = 0x9e3779b9;
        int b = 0x9e3779b9;
        int c = init;
        int k = 0;
        int len;
        int mix[];
        for(len = length; len >= 12; len -= 12) {
            a += url[k + 0] + (url[k + 1] << 8) + (url[k + 2] << 16) + (url[k + 3] << 24);
            b += url[k + 4] + (url[k + 5] << 8) + (url[k + 6] << 16) + (url[k + 7] << 24);
            c += url[k + 8] + (url[k + 9] << 8) + (url[k + 10] << 16) + (url[k + 11] << 24);
            mix = mix(a, b, c);
            a = mix[0];
            b = mix[1];
            c = mix[2];
            k += 12;
        }
        
        c += length;
        switch(len) {
            case 11: // '\013'
                c += url[k + 10] << 24;
                // fall through
                
            case 10: // '\n'
                c += url[k + 9] << 16;
                // fall through
                
            case 9: // '\t'
                c += url[k + 8] << 8;
                // fall through
                
            case 8: // '\b'
                b += url[k + 7] << 24;
                // fall through
                
            case 7: // '\007'
                b += url[k + 6] << 16;
                // fall through
                
            case 6: // '\006'
                b += url[k + 5] << 8;
                // fall through
                
            case 5: // '\005'
                b += url[k + 4];
                // fall through
                
            case 4: // '\004'
                a += url[k + 3] << 24;
                // fall through
                
            case 3: // '\003'
                a += url[k + 2] << 16;
                // fall through
                
            case 2: // '\002'
                a += url[k + 1] << 8;
                // fall through
                
            case 1: // '\001'
                a += url[k + 0];
                // fall through
                
            default:
                mix = mix(a, b, c);
                break;
        }
        return mix[2];
    }
	
}
