package thesis.vector.combination;

import java.io.IOException;
import java.util.ArrayList;

import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.fs.FileSystem;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.io.LongWritable;
import org.apache.hadoop.io.SequenceFile;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.jobcontrol.Job;
import org.apache.mahout.math.RandomAccessSparseVector;
import org.apache.mahout.math.Vector;
import org.apache.mahout.math.VectorWritable;
import org.apache.mahout.utils.vectors.SequenceFileVectorIterable;
import org.apache.mahout.utils.vectors.SequenceFileVectorIterable.SeqFileIterator;

public class VectorCombination {

	private static final double EMPTY_VALUE = 0.0;
	private static final int PAGERANK_VECTOR_DIMENSION = 11;
	
	private static SequenceFile.Writer getSeqFileWriter(String outFile) throws IOException 
	{
		Path path = new Path(outFile);
		Configuration conf = new Configuration();
		FileSystem fs = FileSystem.get(conf);
		// TODO: Make this parameter driven
		SequenceFile.Writer seqWriter = SequenceFile.createWriter(fs, conf, path, LongWritable.class, VectorWritable.class);
		return seqWriter;
	}
	
	private static RandomAccessSparseVector merge(Vector vector1, Vector vector2)
	{
    	int cardinality = vector1.size() + vector2.size();
    	RandomAccessSparseVector newVector = new RandomAccessSparseVector(cardinality);
    	
    	int index = 0;
    	double value = 0.0;
    	for (int i = 0; i < vector1.size(); i++)
    	{
    		value = vector1.getQuick(i);
    		if (value != EMPTY_VALUE)
    			newVector.setQuick(index, value);
    		index++;
    	}
    	for (int i = 0; i < vector2.size(); i++)
    	{
    		value = vector2.getQuick(i);
    		if (value != EMPTY_VALUE)
    			newVector.setQuick(index, value);
    		index++;
    	}
    	
    	newVector.setName(vector1.getName());
		return newVector;
	}
	
	public static void combine2vectors(String inputSeq1, String inputSeq2, String outputSeq) throws IOException
	{
		Path path1 = new Path(inputSeq1);
		JobConf conf1 = new JobConf(Job.class);
		FileSystem fs1 = FileSystem.get(path1.toUri(), conf1);
		SequenceFile.Reader reader1 = new SequenceFile.Reader(fs1, path1, conf1);
		
		Path path2 = new Path(inputSeq2);
		JobConf conf2 = new JobConf(Job.class);
		FileSystem fs2 = FileSystem.get(path2.toUri(), conf2);
		SequenceFile.Reader reader2 = new SequenceFile.Reader(fs2, path2, conf2);
		
        SequenceFileVectorIterable vectorIterable1 = new SequenceFileVectorIterable(reader1, false);
        SequenceFileVectorIterable vectorIterable2 = new SequenceFileVectorIterable(reader2, false);
        
        SeqFileIterator iterator1 = (SeqFileIterator) vectorIterable1.iterator();
        SeqFileIterator iterator2 = (SeqFileIterator) vectorIterable2.iterator();
        
        SequenceFile.Writer writer = getSeqFileWriter(outputSeq);
        
        long i = 0;
        // the size of inputSeq1 and inputSeq2 should be the same
        while (iterator1.hasNext() && iterator2.hasNext())
        {
        	Vector vector1 = iterator1.next();
        	Vector vector2 = iterator2.next();
        	
        	writer.append(new LongWritable(i), new VectorWritable(merge(vector1, vector2)));
        	i++;
        }
        
        writer.close();
	}
	
	public static void combine3vectors(String inputSeq1, String inputSeq2, String inputSeq3, String outputSeq) throws IOException
	{
		// TODO
		Path path1 = new Path(inputSeq1);
		JobConf conf1 = new JobConf(Job.class);
		FileSystem fs1 = FileSystem.get(path1.toUri(), conf1);
		SequenceFile.Reader reader1 = new SequenceFile.Reader(fs1, path1, conf1);
		
		Path path2 = new Path(inputSeq2);
		JobConf conf2 = new JobConf(Job.class);
		FileSystem fs2 = FileSystem.get(path2.toUri(), conf2);
		SequenceFile.Reader reader2 = new SequenceFile.Reader(fs2, path2, conf2);
		
		Path path3 = new Path(inputSeq3);
		JobConf conf3 = new JobConf(Job.class);
		FileSystem fs3 = FileSystem.get(path3.toUri(), conf3);
		SequenceFile.Reader reader3 = new SequenceFile.Reader(fs3, path3, conf3);
		
        SequenceFileVectorIterable vectorIterable1 = new SequenceFileVectorIterable(reader1, false);
        SequenceFileVectorIterable vectorIterable2 = new SequenceFileVectorIterable(reader2, false);
        SequenceFileVectorIterable vectorIterable3 = new SequenceFileVectorIterable(reader3, false);
        
        SeqFileIterator iterator1 = (SeqFileIterator) vectorIterable1.iterator();
        SeqFileIterator iterator2 = (SeqFileIterator) vectorIterable2.iterator();
        SeqFileIterator iterator3 = (SeqFileIterator) vectorIterable3.iterator();
        
        SequenceFile.Writer writer = getSeqFileWriter(outputSeq);
        
        long i = 0;
        // the size of inputSeq1, inputSeq2 and inputSeq3 should be the same
        while (iterator1.hasNext() && iterator2.hasNext() && iterator3.hasNext())
        {
        	Vector vector1 = iterator1.next();
        	Vector vector2 = iterator2.next();
        	Vector vector3 = iterator3.next();
        	writer.append(new LongWritable(i), new VectorWritable(merge(merge(vector1, vector2), vector3)));
        	i++;
        }
        
        writer.close();
		
	}
		
	public static void combine3vectors1pr(String inputSeq1, String inputSeq2, String inputSeq3, String prFile, String outputSeq, boolean isSinglePr) throws IOException, ClassNotFoundException
	{
		Path path1 = new Path(inputSeq1);
		JobConf conf1 = new JobConf(Job.class);
		FileSystem fs1 = FileSystem.get(path1.toUri(), conf1);
		SequenceFile.Reader reader1 = new SequenceFile.Reader(fs1, path1, conf1);
		
		Path path2 = new Path(inputSeq2);
		JobConf conf2 = new JobConf(Job.class);
		FileSystem fs2 = FileSystem.get(path2.toUri(), conf2);
		SequenceFile.Reader reader2 = new SequenceFile.Reader(fs2, path2, conf2);
		
		Path path3 = new Path(inputSeq3);
		JobConf conf3 = new JobConf(Job.class);
		FileSystem fs3 = FileSystem.get(path3.toUri(), conf3);
		SequenceFile.Reader reader3 = new SequenceFile.Reader(fs3, path3, conf3);
		
        SequenceFileVectorIterable vectorIterable1 = new SequenceFileVectorIterable(reader1, false);
        SequenceFileVectorIterable vectorIterable2 = new SequenceFileVectorIterable(reader2, false);
        SequenceFileVectorIterable vectorIterable3 = new SequenceFileVectorIterable(reader3, false);
        
        SeqFileIterator iterator1 = (SeqFileIterator) vectorIterable1.iterator();
        SeqFileIterator iterator2 = (SeqFileIterator) vectorIterable2.iterator();
        SeqFileIterator iterator3 = (SeqFileIterator) vectorIterable3.iterator();

        SequenceFile.Writer writer = getSeqFileWriter(outputSeq);
        ArrayList<String> prList = PageRankSerializer.deserialize(prFile);
        
        int prListIndex = 0;
        if (isSinglePr)
        {
        	while (iterator1.hasNext() && iterator2.hasNext() && iterator3.hasNext())
        	{
        		Vector vector1 = iterator1.next(), vector2 = iterator2.next(), vector3 = iterator3.next();
        		int size1 = vector1.size(), size2 = vector2.size(), size3 = vector3.size();        		
        		int index = 0;
        		
        		RandomAccessSparseVector newVector = new RandomAccessSparseVector(size1 + size2 + size3 + 1);

        		// add first vector
        		for (int i = 0; i < size1; i++)
        		{
        			double value = vector1.getQuick(i);
        			if (value != EMPTY_VALUE)
        			{
        				newVector.setQuick(index, value);
        			}
        			index++;
        		}
        		// add second vector
        		for (int i = 0; i < size2; i++)
        		{
        			double value = vector2.getQuick(i);
        			if (value != EMPTY_VALUE)
        			{
        				newVector.setQuick(index, value);
        			}
        			index++;
        		}
        		// add third vector
        		for (int i = 0; i < size3; i++)
        		{
        			double value = vector3.getQuick(i);
        			if (value != EMPTY_VALUE)
        			{
        				newVector.setQuick(index, value);
        			}
        			index++;
        		}
        		// add page rank
        		double pr = Double.parseDouble(prList.get(prListIndex));
        		if (pr != EMPTY_VALUE)
        		{
        			newVector.setQuick(index, pr);
        		}
        		// set name/lengthSquared
        		newVector.setName(vector1.getName());
        		
        		// write to sequence file
        		writer.append(new LongWritable(prListIndex), new VectorWritable(newVector));
        		prListIndex++;
        	}
        }
        else
        {
        	while (iterator1.hasNext() && iterator2.hasNext() && iterator3.hasNext())
        	{
        		Vector vector1 = iterator1.next(), vector2 = iterator2.next(), vector3 = iterator3.next();
        		int size1 = vector1.size(), size2 = vector2.size(), size3 = vector3.size();
        		int index = 0;
        		RandomAccessSparseVector newVector = new RandomAccessSparseVector(size1 + size2 + size3 + PAGERANK_VECTOR_DIMENSION);
        		
        		// add first vector
        		for (int i = 0; i < size1; i++)
        		{
        			double value = vector1.getQuick(i);
        			if (value != EMPTY_VALUE)
        			{
        				newVector.setQuick(index, value);
        			}
        			index++;
        		}
        		// add second vector
        		for (int i = 0; i < size2; i++)
        		{
        			double value = vector2.getQuick(i);
        			if (value != EMPTY_VALUE)
        			{
        				newVector.setQuick(index, value);
        			}
        			index++;
        		}
        		// add third vector
        		for (int i = 0; i < size3; i++)
        		{
        			double value = vector3.getQuick(i);
        			if (value != EMPTY_VALUE)
        			{
        				newVector.setQuick(index, value);
        			}
        			index++;
        		}
        		// add page rank
        		double pr = Double.parseDouble(prList.get(prListIndex));
        		newVector.setQuick(size1 + size2 + size3 + (int)pr, 1.0);
        		// set name
        		newVector.setName(vector1.getName());
        		// write to sequence file
        		writer.append(new LongWritable(prListIndex), new VectorWritable(newVector));
        		prListIndex++;
        	}
        }
		writer.close();
	}
	
	public static void combine2vectors1pr(String inputSeq1, String inputSeq2, String prFile, String outputSeq, boolean isSinglePr) throws IOException, ClassNotFoundException
	{
		Path path1 = new Path(inputSeq1);
		JobConf conf1 = new JobConf(Job.class);
		FileSystem fs1 = FileSystem.get(path1.toUri(), conf1);
		SequenceFile.Reader reader1 = new SequenceFile.Reader(fs1, path1, conf1);
		
		Path path2 = new Path(inputSeq2);
		JobConf conf2 = new JobConf(Job.class);
		FileSystem fs2 = FileSystem.get(path2.toUri(), conf2);
		SequenceFile.Reader reader2 = new SequenceFile.Reader(fs2, path2, conf2);
		
        SequenceFileVectorIterable vectorIterable1 = new SequenceFileVectorIterable(reader1, false);
        SequenceFileVectorIterable vectorIterable2 = new SequenceFileVectorIterable(reader2, false);
        
        SeqFileIterator iterator1 = (SeqFileIterator) vectorIterable1.iterator();
        SeqFileIterator iterator2 = (SeqFileIterator) vectorIterable2.iterator();

        SequenceFile.Writer writer = getSeqFileWriter(outputSeq);
        ArrayList<String> prList = PageRankSerializer.deserialize(prFile);
        
        int prListIndex = 0;
        if (isSinglePr)
        {
        	while (iterator1.hasNext() && iterator2.hasNext())
        	{
        		Vector vector1 = iterator1.next(), vector2 = iterator2.next();
        		int size1 = vector1.size(), size2 = vector2.size();
        		int index = 0;
        		RandomAccessSparseVector newVector = new RandomAccessSparseVector(size1 + size2 + 1);
        		// add first vector
        		for (int i = 0; i < size1; i++)
        		{
        			double value = vector1.getQuick(i);
        			if (value != EMPTY_VALUE)
        				newVector.setQuick(index, value);
        			index++;
        		}
        		// add second vector
        		for (int i = 0; i < size2; i++)
        		{
        			double value = vector2.getQuick(i);
        			if (value != EMPTY_VALUE)
        				newVector.setQuick(index, value);
        			index++;
        		}
        		// add page rank
        		double pr = Double.parseDouble(prList.get(prListIndex));
        		if (pr != EMPTY_VALUE)
        			newVector.setQuick(index, pr);
        		// write to sequence file
        		writer.append(new LongWritable(prListIndex), new VectorWritable(newVector));
        		prListIndex++;
        	}
        }
        else
        {
        	while (iterator1.hasNext() && iterator2.hasNext())
        	{
        		Vector vector1 = iterator1.next(), vector2 = iterator2.next();
        		int size1 = vector1.size(), size2 = vector2.size();
        		int index = 0;
        		RandomAccessSparseVector newVector = new RandomAccessSparseVector(size1 + size2 + PAGERANK_VECTOR_DIMENSION);
        		// add first vector
        		for (int i = 0; i < size1; i++)
        		{
        			double value = vector1.getQuick(i);
        			if (value != EMPTY_VALUE)
        				newVector.setQuick(index, value);
        			index++;
        		}
        		// add second vector
        		for (int i = 0; i < size2; i++)
        		{
        			double value = vector2.getQuick(i);
        			if (value != EMPTY_VALUE)
        				newVector.setQuick(index, value);
        			index++;
        		}
        		// add page rank
        		double pr = Double.parseDouble(prList.get(prListIndex));
        		newVector.setQuick(size1 + size2 + (int)pr, 1.0);
        		// write to sequence file
        		writer.append(new LongWritable(prListIndex), new VectorWritable(newVector));
        		prListIndex++;
        	}
        }
		writer.close();
	}
	
	public static void combine1vector1pr(String inputSeq1, String prFile, String outputSeq, boolean isSinglePr) throws IOException, ClassNotFoundException
	{
		Path path1 = new Path(inputSeq1);
		JobConf conf1 = new JobConf(Job.class);
		FileSystem fs1 = FileSystem.get(path1.toUri(), conf1);
		SequenceFile.Reader reader1 = new SequenceFile.Reader(fs1, path1, conf1);
		
        SequenceFileVectorIterable vectorIterable1 = new SequenceFileVectorIterable(reader1, false);
        
        SeqFileIterator iterator1 = (SeqFileIterator) vectorIterable1.iterator();
        
        SequenceFile.Writer writer = getSeqFileWriter(outputSeq);
        ArrayList<String> prList = PageRankSerializer.deserialize(prFile);
        
        int prListIndex = 0;
        if (isSinglePr)
        {
        	//int index = 0;
        	while (iterator1.hasNext())
        	{
        		int index = 0;
        		Vector vector1 = iterator1.next();
        		int size = vector1.size();
        		RandomAccessSparseVector newVector = new RandomAccessSparseVector(size + 1);

        		// add vector
        		for (int i = 0; i < size; i++)
        		{
        			double value = vector1.getQuick(i);
        			if (value != EMPTY_VALUE)
        				newVector.setQuick(index, value);
        			index++;
        		}
        		// add page rank
        		double pr = Double.valueOf(prList.get(prListIndex));
        		if (pr != EMPTY_VALUE)
        			newVector.setQuick(index, pr);
        		
        		// set name
        		newVector.setName(vector1.getName());
        		// write the new vector into sequenceFile
        		writer.append(new LongWritable(prListIndex), new VectorWritable(newVector));
        		prListIndex++;
        	}
        }
        else
        {
        	//int index = 0;
        	while (iterator1.hasNext())
        	{
        		int index = 0;
        		Vector vector1 = iterator1.next();
        		int size = vector1.size();
        		RandomAccessSparseVector newVector = new RandomAccessSparseVector(size + PAGERANK_VECTOR_DIMENSION);
        		// add vector
        		for (int i = 0; i < size; i++)
        		{
        			double value = vector1.getQuick(i);
        			if (value != EMPTY_VALUE)
        				newVector.setQuick(index, value);
        			index++;
        		}
        		// add page rank
        		double pr = Double.valueOf(prList.get(prListIndex));
        		newVector.setQuick(size + (int)pr, 1.0);
        		
        		// set name
        		newVector.setName(vector1.getName());
        		// write to sequence file
        		writer.append(new LongWritable(prListIndex), new VectorWritable(newVector));
        		prListIndex++;
        	}
        }
        writer.close();
	}

}
